# Copyright 2014-2015 Neil Freeman contact@fakeisthenewreal.org
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from twitter_bot_utils import helpers
import wordfilter
import re


def reply_checker(tweet):
    try:
        if tweet.in_reply_to_user_id:
            return False

    except AttributeError:
        try:
            if tweet.get('in_reply_to_user_id'):
                return False

        except AttributeError:

            try:
                if "@" == tweet[0]:
                    return False

            except AttributeError:
                pass

    return True


def rt_checker(tweet):
    try:
        if tweet.retweeted:
            return False

    except AttributeError:
        try:
            if tweet.get('retweeted_status'):
                return False

        except AttributeError:

            try:
                if "RT" in tweet[:2]:
                    return False

            except AttributeError:
                pass

    return True


def wf_checker(tweet):
    try:
        if wordfilter.blacklisted(tweet.text):
            return False

    except AttributeError:
        try:
            if wordfilter.blacklisted(tweet['text']):
                return False

        except (KeyError, TypeError):
            pass

    return True

def construct_tweet_checker(no_retweets=False, no_replies=False, no_badwords=True):
    '''Returns a tweet checker'''
    checks = []

    if no_retweets:
        checks.append(rt_checker)

    if no_replies:
        checks.append(reply_checker)

    if no_badwords:
        checks.append(wf_checker)

    def checker(tweet):
        for check in checks:
            if not check(tweet):
                return False
        return True

    return checker


def construct_tweet_filter(no_mentions=False, no_urls=False, no_media=False, no_hashtags=False, no_symbols=False):
    '''returns a filter for tweet text'''

    entitytypes = []

    if no_mentions:
        entitytypes.append('user_mentions')

    if no_hashtags:
        entitytypes.append('hashtags')

    if no_urls:
        entitytypes.append('urls')

    if no_media:
        entitytypes.append('media')

    if no_symbols:
        entitytypes.append('symbols')

    def filterer(tweet):
        # ignore strings
        if isinstance(tweet, str):
            text = tweet

        else:
            text = helpers.remove_entities(tweet, entitytypes)

        # Older tweets don't have entities
        if no_urls:
            # regex stolen from http://stackoverflow.com/questions/6883049/regex-to-find-urls-in-string-in-python
            text = re.sub(r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+", '', text)

        if no_mentions:
            text = re.sub(r'@\w+', '', text)

        if no_hashtags:
            text = re.sub(r'#\w+', '', text)

        if no_symbols:
            text = re.sub(r'\$[a-zA-Z]+', '', text)

        return text

    return filterer



# Copyright 2014-2015 Neil Freeman contact@fakeisthenewreal.org
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, print_function
import os
import re
import logging
import Levenshtein
from cobe import scoring
from cobe.brain import Brain
import twitter_bot_utils as tbu
from wordfilter import Wordfilter
from . import checking


class Twitter_markov(object):

    """Posts markov-generated text to twitter"""

    default_brain = None
    _recently_tweeted = []

    def __init__(self, screen_name, brains=None, **kwargs):

        self.screen_name = screen_name

        self.api = kwargs.get('api', tbu.api.API(screen_name, **kwargs))

        self.config = self.api.config

        self.logger = logging.getLogger(screen_name)

        try:
            if isinstance(brains, str):
                brains = [brains]

            if not brains:
                brains = [self.config.get('brain')] + self.config.get('brains', [])
                brains = [b for b in brains if b is not None]

            self.brains = self._setup_brains(brains)

        except (IOError, IndexError, RuntimeError) as e:
            self.logger.error('Feed me brains: unable to find any brains!')
            raise e

        self.logger.debug('Brains: {0}'.format(list(self.brains.keys())))

        self.dry_run = kwargs.get('dry_run', False)

        self.wordfilter = Wordfilter()
        self.wordfilter.add_words(self.config.get('blacklist', []))

        if kwargs.get('learn', True):
            self.learn_parent()

    def _setup_brains(self, brains):
        self.logger.debug('setting up brains')
        out = dict()

        try:
            for pth in brains:
                brainpath = os.path.expanduser(pth)
                name = os.path.basename(brainpath).replace('.brain', '')

                if not os.path.exists(brainpath):
                    raise IOError("Brain file '{0}' missing".format(brainpath))

                out[name] = Brain(brainpath)
                out[name].scorer.add_scorer(2.0, scoring.LengthScorer())

        except AttributeError as e:
            self.logger.error(e)
            self.logger.error("Probably couldn't find the brain file.")
            raise e

        except IOError as e:
            self.logger.error(e)
            self.logger.error(brains)
            raise e

        self.default_brain = os.path.basename(brains[0]).replace('.brain', '')

        return out

    @property
    def recently_tweeted(self):
        if len(self._recently_tweeted) == 0:
            recent_tweets = self.api.user_timeline(self.screen_name, count=self.config.get('checkback', 20))
            self._recently_tweeted = [x.text for x in recent_tweets]

        return self._recently_tweeted

    def check_tweet(self, text):
        text = text.strip().lower()

        if len(text) == 0:
            self.logger.info("Rejected (empty)")
            return False

        if self.wordfilter.blacklisted(text):
            self.logger.info("Rejected (blacklisted)")
            return False

        for line in self.recently_tweeted:
            if text in line.strip().lower():
                self.logger.info("Rejected (Identical)")
                return False

            if Levenshtein.ratio(re.sub(r'\W+', '', text), re.sub(r'\W+', '', line.lower())) >= 0.70:
                self.logger.info("Rejected (Levenshtein.ratio)")
                return False

        return True

    def reply_all(self, brainname=None):
        mentions = self.api.mentions_timeline(since_id=self.api.last_reply)
        self.logger.debug('{0} mentions found'.format(len(mentions)))

        for status in mentions:
            self.reply(status, brainname)

    def reply(self, status, brainname=None):
        self.logger.debug('Replying to a mention')

        if status.user.screen_name == self.screen_name:
            self.logger.debug('Not replying to self')
            return

        catalyst = tbu.helpers.format_status(status)
        text = self.compose(catalyst, brainname, max_len=138 - len(status.user.screen_name))

        reply = '@' + status.user.screen_name + ' ' + text

        self.logger.info(reply)
        self._update(reply, in_reply=status.id_str)

    def tweet(self, catalyst='', brainname=None):
        self.logger.debug('tweeting')

        text = self.compose(catalyst, brainname)

        self.logger.info(text)
        self._update(text)

    def _update(self, tweet, in_reply=None):
        if not self.dry_run:
            self.api.update_status(status=tweet, in_reply_to_status_id=in_reply)

    def compose(self, catalyst='', brainname=None, max_len=140):
        '''Format a tweet with a reply from brainname'''

        max_len = min(140, max_len)

        brainname = brainname or self.default_brain
        brain = self.brains[brainname]

        reply = brain.reply(catalyst, max_len=max_len)

        self.logger.debug('input> ' + catalyst)
        self.logger.debug('reply> ' + reply)

        if len(reply) <= 140:
            return reply

        else:
            self.logger.debug('Tweet was too long, trying again')
            return self.compose(catalyst, brainname, max_len)

    def learn_parent(self, brainname=None):
        parent = self.config.get('parent')

        last_tweet = self.api.last_tweet

        if not parent or not last_tweet:
            return

        tweet_filter = checking.construct_tweet_filter(
            no_mentions=self.config.get('filter_mentions'),
            no_hashtags=self.config.get('filter_hashtags'),
            no_urls=self.config.get('filter_urls'),
            no_media=self.config.get('filter_media'),
            no_symbols=self.config.get('filter_symbols')
        )

        tweet_checker = checking.construct_tweet_checker(
            no_badwords=self.config.get('filter_parent_badwords', True),
            no_retweets=self.config.get('no_retweets'),
            no_replies=self.config.get('no_replies')
            )

        tweets = self.api.user_timeline(parent, since_id=last_tweet)

        brain = brainname or self.default_brain

        for status in tweets:
            if not tweet_checker(status):
                continue

            text = tweet_filter(status)

            text = tbu.helpers.format_text(text)

            self.brains[brain].learn(text)



# Copyright 2014-2015 Neil Freeman contact@fakeisthenewreal.org
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from cobe.brain import Brain
from . import checking
from twitter_bot_utils import helpers, archive


def tweet_generator(archivepath, **kwargs):
    if kwargs.get('txt'):
        generator = archive.read_text(archivepath)
    else:
        generator = archive.read_json(archivepath)

    cool_tweet = checking.construct_tweet_checker(
        kwargs.get('no_retweets', False), kwargs.get('no_replies', False))

    tweet_filter = checking.construct_tweet_filter(
        no_mentions=kwargs.get('no_mentions', False),
        no_urls=kwargs.get('no_urls', False),
        no_media=kwargs.get('no_media', False),
        no_hashtags=kwargs.get('no_hashtags', False),
        no_symbols=kwargs.get('no_symbols', False)
    )

    for status in generator:
        if not cool_tweet(status):
            continue

        text = tweet_filter(status)
        text = helpers.format_text(text)
        yield text


def learn(archivepath, brain, **kwargs):
    # start brain. Batch saves us from lots of I/O
    brain = Brain(brain)
    brain.set_stemmer(kwargs.get('language', 'english'))

    brain.start_batch_learning()

    tweets = tweet_generator(archivepath, **kwargs)
    count = 0

    for text in tweets:
        count = count + 1
        brain.learn(text)

    brain.stop_batch_learning()

    return count



# Copyright 2014-2015 Neil Freeman contact@fakeisthenewreal.org
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, print_function
import os
import sys
import logging
import argparse
import twitter_bot_utils.args as tbu
from .learn import learn
from .twitter_markov import Twitter_markov
from . import __version__


def main():
    parser = argparse.ArgumentParser('twittermarkov', description='Tweet with a markov bot, or teach it from a twitter archive.')

    parser.add_argument('-V', '--version', action='version', version="%(prog)s " + __version__)

    subparsers = parser.add_subparsers()
    tweet = subparsers.add_parser(
        'tweet', parents=[tbu.parent()],
        description='Post markov chain ("ebooks") tweets to Twitter', usage='%(prog)s [options] SCREEN_NAME')

    tweet.add_argument('-r', '--reply', action='store_const', const='reply', dest='action', help='tweet responses to recent mentions')
    tweet.add_argument('--brain', dest='brains', metavar='BRAIN', type=str, help='cobe .brain file')
    tweet.add_argument('--no-learn', dest='learn', action='store_false', help='skip learning')
    tweet.add_argument('screen_name', type=str, metavar='SCREEN_NAME', help='User who will be tweeting')
    tweet.set_defaults(func=tweet_func, action='tweet')

    learnparser = subparsers.add_parser(
        'learn', description='Teach a Cobe brain the contents of a Twitter archive',
        usage="%(prog)s [options] ARCHIVEPATH NEWBRAIN")

    learnparser.add_argument('--no-replies', action='store_true', help='skip replies')
    learnparser.add_argument('--no-retweets', action='store_true', help='skip retweets')
    learnparser.add_argument('--no-urls', action='store_true', help='Filter out urls')
    learnparser.add_argument('--no-media', action='store_true', help='filter out media')
    learnparser.add_argument('--no-hashtags', action='store_true', help='filter out hashtags')
    learnparser.add_argument(
        '--language', type=str, default='english', help='language. Defaults to English')
    learnparser.add_argument(
        '--txt', action='store_true', help='Read from a text file, one tweet per line')
    learnparser.add_argument('-q', '--quiet', action='store_true', help='run quietly')
    learnparser.add_argument('archive', type=str, metavar='ARCHIVEPATH',
                             default=os.getcwd(), help='top-level folder of twitter archive')
    learnparser.add_argument('brain', type=str, metavar='NEWBRAIN', help='brain file to create')
    learnparser.set_defaults(func=learn_func)

    args = parser.parse_args()

    argdict = vars(args)
    del argdict['func']

    args.func(argdict)


def tweet_func(args):
    tbu.add_logger(args['screen_name'], args['verbose'])
    logger = logging.getLogger(args['screen_name'])

    tm = Twitter_markov(**args)

    if args['tweet']:
        logger.debug('tweeting...')
        tm.tweet()

    if args['reply']:
        logger.debug('replying to all')
        tm.reply_all()


def learn_func(args):
    if not args['quiet']:
        print("Reading from " + args['archive'], file=sys.stderr)
        print("Teaching " + args['brain'], file=sys.stderr)

    if args['brain'][-6:] == '.brain':
        brainpath = args['brain']
    else:
        brainpath = args['brain'] + '.brain'

    count = learn(args['archive'], brainpath, **args)

    if not args['quiet']:
        print("Taught {0} tweets".format(count))


if __name__ == '__main__':
    main()
